# pio_buzzer_1.py - 12.09.2021
#
# Zwei Staemachines arbeiten zusammen und bilden eeinen Intervall-
# Summer:
# sm1 erezugt das Rechtecksignal für den Summer
# sm2 schaltet den Summer ein- und aus.
# Die Kommunikation zwischen dem Taktgeber und dem Tongenerator
# erfolgt über einen Interrupt.
# Beide Staremachines müssen auf dem gleichen PIO-Block laufen.
# Die Interruprs 0 n- 7 sind einsetzbar.
#
# Der Grundtakt ist 1 Hz, kann mit "speed" erhöht werden.
# Weniger als 1 Hz sind nicht möglich ohne Änderung des Codes.
# TX_FIFO ist um den RX_FIFO erweitert (auf 8 Words).

from rp2 import PIO, StateMachine, asm_pio
from machine import Pin

class Buzzer(object):
    def __init__(self, output_pin, buzzer_freq=3250, speed=4):
        speed = 1 if speed < 1 else speed
        self.sm0 = StateMachine(0, self.buzzer, freq=buzzer_freq*4, set_base=Pin(output_pin))
        self.sm1 = StateMachine(1, self.cycle, freq=int(round(2000 * speed)) )
        self.sm0.active(1)
        self.sm1.active(1)
        
    def beep(self, cnt=3):
        cnt = 1 if cnt < 1 else cnt
        if self.tx_free():
            self.sm1.put(cnt - 1)
        else:
            print("Fifo full !")
    
    def tx_free(self):
        # Anzahl der freien Words im RX-Fifo
        return 8 - self.sm1.tx_fifo()
    
    @staticmethod
    @asm_pio(set_init=PIO.OUT_LOW)
    # Hier wird der Ton generiert = 1/4 des F_PIO-Taktes
    def buzzer():
        IRQ = 4                 #    IRQ-Nummer als Variable
        
        wrap_target()  
        set(pins, 0)            #  1 4 Takte für 1x (Pulse + Pause)
                                #    warten, falls IRQ x gesetzt ist
        wait (0, irq, IRQ)      #  2 0 = das Flag wird nicht gelöscht
        set(pins, 1)        [1] #  3  
        wrap()
        
    @staticmethod
    @asm_pio(out_shiftdir=PIO.SHIFT_RIGHT, fifo_join=PIO.JOIN_TX)
    # hier werden Pulse und Pause generiert,
    # die den Buzzer ein- und ausschalten.

    def cycle():
        IRQ = 4                  #    IRQ-Nummer als Variable
        wrap_target()
        pull(block)              #  1 blockend warten auf eine Eingabe
        mov(y, osr)              #  2 Anzahl der Beeps nacch Y
        
        label("start")           #    Beginn der inneren Schleife
        irq (clear, IRQ)         #  3 IRQ löschen - 1000 Takte
        set(x, 30)           [6] #  4 1 + (6 + 1) + ((30 + 1) * (31 + 1))
                              
        label("on")              #    Schleife für Pulse
        jmp (x_dec, "on")   [31] #  5

        irq(IRQ)                 #  6 IRQ setzen / 1000 Takte
        set(x, 30)           [6] #  7 1 + (6 + 1) + (30 + 1) * (31 + 1) 
                              
        label("off")             #    Schleife für Pause 31 * 32
        jmp(x_dec, "off")   [31] #  8
        
        jmp(y_dec, "start")      #  9 Wiederholen ?
        
        set(x, 31)          [31] # 10 eine zusätzliche Pause am Ende einfügen
                                 #    32 + 2048 = 2080 Takte
        label("wait")            #    31 + 1 + (31 + 1) * (31 + 1 + 31 + 1)
        nop()               [31] # 11
        jmp(x_dec, "wait")  [31] # 12

        wrap() 

if __name__ == "__main__":
    
    buzzer_pin = 19
    buzz = Buzzer(output_pin=buzzer_pin, speed=3)
    
    for i in range(3):
        buzz.beep(2 + i)
    
    print("\ndone")
