# pio_counter_sm1_5T.py -12.09.2021
#
# Counter für 1 Statemachine bei 5 PIO-Takten pro Count
#
# Die kleinste Zeiteinheit für die Messung ist 5/F_PIO = 40ns bei 125 MHz
# Bei F_PIO 5 MHz 1.0us, der Counter zählt bis 2**16
# 5 PIO-Takte sind erforderlich (Reduktion auf 3 ist machbar),
# um eine Pulse-Pause-Folge abzutasten.
# Verwendet 1 einzige Statemchine zum Zählen.
# Der Pico muss den FIFO schnell genug auslesen - das geht bis ca. 20 KHz
#
# Der Ruhepegel ist high.
# kann zum Testen an TX einer Serielle Schnittstelle angeschlossen werden.

import time
import uarray as arr
from rp2 import PIO, StateMachine, asm_pio
from machine import Pin, Timer

# Erzeugt das Testsignal von x Pulse-Pause Folgen -----------------  
class Pulse_Emitter(object):
    def __init__(self, statemachine, tx_pin, pulse_freq = 1000):
       self.sm = StateMachine(statemachine, self.pulse_train,
                        freq=pulse_freq*10,
                        set_base=Pin(tx_pin))
       self.sm.restart()
       
    def start_sending(self, pulse_cnt):
        self.sm.active(1)
        self.sm.put(pulse_cnt - 1)
        
    @staticmethod  
    @asm_pio(set_init=PIO.OUT_HIGH,
                  in_shiftdir=PIO.SHIFT_RIGHT)
    def pulse_train():
        
        wrap_target()
        pull(block)            #  1 warten auf Input/Anzahl der Pulse
        mov(x, osr)            #  2 nach x zum Downcounten

        label("next")          #    pulse + pause = 10 Takte (4 + 6)
        set(pins, 0)      [4]  #  3
        set (pins, 1)     [3]  #  4
        jmp(x_dec, "next")     #  5

        wrap()

# - Der Counter ----------------------------------------
@asm_pio(in_shiftdir=PIO.SHIFT_LEFT,
         sideset_init=rp2.PIO.OUT_HIGH,
         fifo_join=PIO.JOIN_RX )

def pulse_counter():
                           #    16-Bit DownCounter 0x0000_FFFF vorbereiten
    mov(osr, invert(null)) #  1 0xFFFF_FFFF in OSR
    in_(osr, 16)           #  2 0x0000_FFFF in ISR
    mov(osr, isr)          #  3 ISR kopieren nach OSR
    mov(x, osr)            #  4 OSR nach x kopieren
            
    #pull()                #    alternativ den Counter via put() importieren
    #mov(x, osr)
    
    label("start")
    
    wait(0, pin, 0)        #  5 Ausgangslevel ist high, jetzt
                           #  6 warten auf eine fallende Flanke
    wrap_target() 
    
                           #    decrementieren, >0, pin  low, weiter  3 + [2]
                           #    decrementieren, >0, pin high,   push  2 + 2 + [1]
                           #    decrementieren,  0, pin  low, weiter  3 + [2] 
    label("low")
    jmp(x_dec, "low_1")    #  7 decrementieren und immer (!) weiter zum nächsten Step                   
    label("low_1")
    
    jmp(pin, "low_end")    #  8 der Pin ist high
    jmp("low")         [2] #  9 der Pin ist weiter low
    
    label("low_end")
    in_(x, 16)             # 10 x ist nicht invertiert - 3 Takte
    mov(x, osr)   [1]
                           #    decrementieren >0, pin high, weiter     3 + [2]
                           #    decrementieren >0, pin  low, speichern  2 + 2 + [1]
                           #    decrementieren  0, pin high, weiter     3 + [2]              
    label("high")
    jmp(x_dec, "high_1")   # 11 Zähler decrementieren, 
    jmp("train_end")       # 12 wenn der Counter überläuft, dann beenden
    label("high_1")        #    hier gehts unabhängig vom Counter Weiter 
    
    jmp(pin, "high_2")     # 13 der Pin ist weiter high
    jmp("high_end")        # 14 der Pin ist low
    
    label("high_2")
    jmp("high")        [2] # 15 Wartetakte einlegen
    
    label("high_end")
    in_(x, 16)            # 16 16 Bit speichern, x ist nicht invertiert
    mov(x, osr)           # 17 Counter neu laden und weiter in low
     
    push()                # 18
    wrap()
    
    label("train_end")    # 19 der Counter ist bei high-level
    in_(x, 16)            # 20 abgelaufen - wird als Ende interpretiert
    push()
    irq(rel(0))           # 21
    
    jmp("start")          # 22 Ende, warten auf ene neur Falling Edge
    
    
class Counter(object):
    """
    Zählt die Dauer von Pulse und Pause in der Einheit 5/counter_freq
    Es werden per default 8 Pulse/Pause Folgen gemessen (aus FIFo),
    das ist soweit nicht zeitkritisch
    Wenn max_cnt > 8, dann können bis ca. 19.000 Words/sec gelesen werden.
    (das Prüfen auf Oberlauf kostet eben ...).
    Wenn max_cnt = None, dann wird ein Überlauf des Buffers nicht geprüft,
    es können dann bis ca. 21.000 Words/sec gelesen werden.
    """
    def __init__(self, statemachine, rx_pin, buffer_size=64, counter_freq=1_000_000):
        self.sm = StateMachine(statemachine, pulse_counter, freq=counter_freq,
            in_base=Pin(rx_pin, Pin.IN, Pin.PULL_UP),
            jmp_pin=Pin(rx_pin, Pin.IN, Pin.PULL_UP),
            sideset_base=Pin(rx_pin + 1))
        self.buffer_size = buffer_size
        self.buffer = arr.array("L", [0] * buffer_size)
        self.timer = Timer()
        self.F_PIO = counter_freq
        self.do_read = True
        self.idx = 0
        self.sm.restart()
        self.sm.irq(self.irq_handler)
        self.sm.active(1)
        
        print("\nF_PIO: {} Hz, resolution: {} Nanosec.".format(self.F_PIO, 5 * 10**9 / self.F_PIO))
                   
    def irq_handler(self, _):
        # Wird ausgeführt nach Ablauf des Timers
        # und wenn Ende der Übertragung erkannt (X-Register = 0)
        self.do_read = False
        self.sm.active(0)
           
    def start_read_blocking(self, max_waittime=1000,check_index=True):
        self.timer.init(mode=Timer.ONE_SHOT,
                        period=max_waittime,
                        callback=self.irq_handler)
        if not check_index:
            print("Buffer_size: {}, Index not checked !".format(self.buffer_size))
            # Buffer Überlauf nicht abfangen
            # beenden, wenn der Timer abgelaufen ist
            while self.do_read:
                if self.sm.rx_fifo():
                    self.buffer[self.idx] = self.sm.get()
                    self.idx += 1         
        else:
            # Buffer-Überlauf verhindern und das Einlesen
            # ohne Warten auf den Timer beenden
            # max_cnt = self.buffer_size if max_cnt > self.buffer_size else max_cnt
            while self.do_read:
                if self.sm.rx_fifo():
                    self.buffer[self.idx] = self.sm.get()
                    self.idx += 1
                    if self.idx >= self.buffer_size:
                        print("- buffer overflow at idx: {} !".format(self.idx))
                        self.timer.deinit()
                        self.do_read = False
                        break
        self.sm.active(0)
      
    def print(self):
        for i in range(self.idx):
            pulse = 0xFFFF - (self.buffer[i] // 256**2)
            pause = 0xFFFF - (self.buffer[i] & 0xFFFF)
            print("# {:03}: {:5} {:5}".format(i, pulse, pause))
            if pause == 0:
                break


if __name__ == "__main__":
    
    # --- Einstellungen -------------------
    
    rx_pin = 17          
    F_PIO = 5_000_000  

    # --------------------------------------
    
    # -- Counter starten
    sm = Counter(0, rx_pin=rx_pin, counter_freq=F_PIO, buffer_size=64)
    # aktivieren bevor der Sender aktiv wird, sonst geht der erste Takt verloren
    
    # Die Daten aus dem FIFO abholen - blocking !!
    # Wird beendet, wenn - der Timer abgelaufen ist oder
    #                    - das Ende einer Übertragung erkannt wird
    #                      (d.h. der 16-Bit Counter überlaäuft !)
    try:
        sm.start_read_blocking(max_waittime = 10_000, check_index=True)
    except KeyboardInterrupt:
        pass
    sm.print()
    print("\ndone.")
