# pi0_pwm.py - 12.09.2921
#
# PWM via PIO - alle Parameter sind einstellbar:
# - pwm_max_count - wieviele Bits
# - pwm_freq      - Wiederholungen pro Sekunde
# - pwm_steps     - wieviele Zwischenschritte zwischen 0 und pwm_max_count
# - pwm_speed     - wie oft pro Sekunde soll der Compare-Wert aktualisiert werden
#
# Die Compare-Werte werden berechnet:
# - factor = pow( pwm_max_count, (1 / (pwm_steps - 1)) )
# - (n + 1) = n * factor  (factor ca. 1.003
#
#    Eine Tabelle kann so angelegt werden:
#
#    import uarray as arr
#    pwm_values = arr.array("H", [0] * pwm_steps)
#    up_count = up_counter()
#    for i in range(pwm_steps):
#        pwm_values[i] = next(up_count)
#    print(pwm_values)
    
from machine import Pin
from rp2 import PIO, StateMachine, asm_pio
from time import sleep
from math import pow

class PIO_PWM:
    def __init__(self, statemachine, output_pin, pwm_max_count, pio_freq):
        self.sm = StateMachine(statemachine, self.pwm, freq=pio_freq,
                                sideset_base=Pin(output_pin))
        self.sm.put(pwm_max_count)
        self.sm.exec("pull()")
        self.sm.exec("mov(isr, osr)")
        self.sm.active(1)

    def set(self, value):
        self.sm.put(value)
        # print(value)

    #staticmethod
    @asm_pio(sideset_init=PIO.OUT_LOW)
    def pwm():
        wrap_target()
        
        pull(noblock) .side(0)       #  1
        mov(x, osr)                  #  2 Max PWM
        mov(y, isr)                  #  3 Compare-Wert
        label("pwmloop")
        jmp(x_not_y, "skip")         #  4
        nop()         .side(1)       #  5
        label("skip")
        jmp(y_dec, "pwmloop")        #  6
        
        wrap()

if __name__ == "__main__":
    
    # -------------------------------------------------------------------
    
    pwm_max_count = 2** 16 - 1 # maximaler Wert
    pwm_freq  = 200            # Wiedehrolfrequenz in Hz -
    pwm_steps = 30            #  wieviele Zwischenschritte zwischen 0 und pwm_max_count
    pwm_speed = 30             # wieviele Compare-Werte pro Sekunde für die Demo
    
    led_pin = 0
    
    # --------------------------------------------------------------------
    
    pio_freq = pwm_max_count * pwm_freq
    factor = pow(pwm_max_count, 1 / (pwm_steps - 1) )
    # print(pio_freq)
    # print(factor)
    
    # -------------------------------------------------------------------

    def up_counter():
        # Generator, zählt von 0 bis pwm_max_count
        pwm = 0
        for _ in range(pwm_steps):
            pwm *= factor
            yield int(round(pwm, 0)) 
            if pwm == 0:
               pwm = 1

    def down_counter():
        # Generator, zählt von pwm_max_count bis 0
        pwm = pwm_max_count
        for i in range(pwm_steps):
            yield int(round(pwm, 0)) 
            pwm /= factor
            if i >= pwm_steps-2:
               pwm = 0
    
    # -------------------------------------------------------------------
    
    pwm = PIO_PWM(statemachine=0,
                  output_pin=led_pin,
                  pwm_max_count=pwm_max_count,
                  pio_freq=pio_freq)                  
    
    print("Ctrl-C to stop !\n")
    
    while True:
        try:
            up_count = up_counter()
            print("up_counting", end = " ")
            for _ in range(pwm_steps):
                pwm.set(next(up_count))
                sleep(1 / pwm_speed)
            print("- top")
            
            down_count = down_counter()
            print("down_counting", end = " ")
            for _ in range(pwm_steps):
                pwm.set(next(down_count))
                sleep(1 / pwm_speed)
            print("- bottom")
            
        except KeyboardInterrupt:
            break
    
    print("\n\ndone")
