# pio_uart_rx_1.py - 12.09.2021
#
# Liest immer 1 einzelnes Byte an der Seriellen Schnittstelle ein
# und löst dann einen Interrupt aus.
# Der Interrupt-Handler incrementiert einen Byte-Counter, main() muss die
# Bytes rechtzeitig abholen, damit der Receiver nicht blockt.

from machine import Pin, lightsleep
from rp2 import PIO, StateMachine, asm_pio
import uarray as arr
import time

# --- Receiver ----------------------------------------------------------
#
# Diese Version prüft das Startbit 2x, das Stopbit 1x.
# Die Datenbits werden jeweils 1x in der Bitmitte abgetastet.
# PIO_Takt = Baud * 9   
@rp2.asm_pio( in_shiftdir=rp2.PIO.SHIFT_RIGHT,
        fifo_join=rp2.PIO.JOIN_RX )

def uart_rx():
    label("restart")    
    wrap_target()
    
    wait(0, pin, 0)           #  1 Warten auf eine fallende Flankem, dann noch 2 Takte warten
    mov(isr, null)            #  2 Bei einem Fehler Sicherheitshalber das ISR zurücksetzen
    set(y, 7)                 #  3 den Counter setzen
                              #    Das Signal ist mindestens 3 Takte gesetzt
    jmp(pin, "restart")   [2] #  4 Fehler, weil der 4. Takt das Startbit nicht mehr low 
    jmp(pin, "restart")   [7] #  5 Fehler, weil der 6. Takt das Startbit nicht mehr low 
                              #    warten bis die Bit-Mitte des 1. Bits erreicht ist
                              #    dh. vom Start aus 9 + 5  = 14 Takte um sind
                              #    1 + 1 + 1 + (1 + 2) + (1 + 7) = 14
    label("nextbit")
                              #   
    in_(pins, 1)          [7] #  6 Den Pegel einlesen und nach
    jmp(y_dec, "nextbit")     #  7 weiteren 9 Takten das nächste Bit einlesen

    jmp(pin, "stop_bit_1")    #  8 Das Stopbit muss high, sein 
    jmp("restart")            #    nicht gesetzt, Fehler
    label("stop_bit_1")
    jmp(pin, "byte_end")      #  9 Das Stopbit muss high sein, 
    jmp("restart")            #    nicht gesetzt, Fehler                         
    
    label("byte_end")

    push()                    # 10 die empfangenen Bits ins Fifo
    irq(rel(0))               # 11 Interrupt auslösen
    wrap()

class UART_RX(object):
    # Incrementiert nach dem Empfang per IRQ nur einen Counter 
    # Aus dem Fifo ausgelesen werden die Daten später
    def __init__(self, statemachine, rx_pin, baud=9600):
        self.sm = StateMachine(statemachine, uart_rx,
                in_base=Pin(rx_pin, Pin.IN, Pin.PULL_UP),
                jmp_pin=Pin(rx_pin, Pin.IN, Pin.PULL_UP),
                freq=baud * 9)
        
        self.sm.irq(self.irq_handler)
        self.sm.restart()
        self.rx_cnt = 0

    def activate(self, state=1):
        self.sm.active(1 if state else 0)
        
    def irq_handler(self, _):
        # Bei jedem Interrupt wurde genau 1 Byte empfangen
        self.rx_cnt += 1
        
    def received(self):
        # Die Anzahl der wartenden Bytes per rx_fifo() ermitteln
        # return self.sm.rx_fifo()
        return self.rx_cnt
    
    def get_data(self):
        # jeweils 1 Byte aus dem Fifo zurücliefern
        self.rx_cnt -= 1
        return self.sm.get(None, 24)

if __name__ == "__main__":
    
    # --- Einstellungen für die Tests ----------------------------
    
    BAUD = 9600 
    rx_pin = 17        
       
    # ---------------------------------------------------------

    rx = UART_RX(statemachine=0, rx_pin=rx_pin, baud=BAUD)
    rx.activate()

    # Buffer zur Aufnahme der erwarteten Daten
    rx_buffer = arr.array("B", [0] * 200)
    rx_cnt = 0

    print("Waiting for UART-Data ... Ctrl-C to stop + print.\n")
    
    try:
        while True:
            #lightsleep(1)
  
            while rx.received():
                # jeweils ein Byte auslesen               
                rx_buffer[rx_cnt] = rx.get_data()
                rx_cnt += 1
                               
    except KeyboardInterrupt:
        print(rx_buffer[:rx_cnt].decode())
        print("Bytes received: ", rx_cnt)
    
    rx.activate(0)
    print("\ndone")
