# pio_wave_asym.py - 12.09.2021
#
# Erzeugt eine Folge von asymmetrischen Impulsen, bei denen
# - die Anzahl der Waves sowie die
# - die Dauer von Pulse und Pause einstellbar sind
#   Zeiteinheit von  2ms (freq_factor = 0.5)
#           default  1ms (freq_factor = 1)
#               bis 20ns (freq_factor = 50_000) 

import time
from rp2 import PIO, StateMachine, asm_pio
from machine import Pin

@asm_pio(sideset_init=PIO.OUT_LOW)
def wave_train_asym():
    
    label("start")

    pull()                     #   1    Warten auf den Start
    mov(y, osr)                #   2 1. Wert nach Y für Anzahl der Waves
    pull()                     #   3
    mov(isr, osr)              #   4 2. Wert nach ISR für Dauer Pulse
    pull()                     #   5 3. Wert in OSR für Dauer Pause
    
    wrap_target()

    jmp(y_dec, "go")           #   6 Beenden, wenn y == 0 ist
    jmp("start")               #   7 dann von vorne beginnen
    
    label("go")
    mov(x, osr)   .side(1)     #   8 Dauer für Pulse laden

    label("wait_on_end")
    jmp(x_dec, "wait_on")      #   9
                                
    nop()                  [2] #  10 1 + 1 + 2 + 1 = 5 Takte 
    mov(x, isr)   .side(0)     #  11 Dauer für Pause laden
                               #     weiter in low
    label("wait_off_end")
    jmp(x_dec, "wait_off")     #  12
    nop()                  [1] #  13 1 + 2  Takte
                               #       + 2 Takte nach wrap
    wrap()                     

    label("wait_on")           #     Warteschleife für high
    jmp("wait_on_end")     [3] #  14 1 + 1 + 3 = 5 Takte

    label("wait_off")          #     Warteschleife für low
    jmp("wait_off_end")    [3] #  15 1 + 1 + 3 = 5 Takte 
    

class Wave_Train(object):
    def __init__(self, output_pin, freq_factor = 1):
        if freq_factor < 0.4 or freq_factor > 50_000:
            print("freq_factor out of range: {}".format(freq_factor))
        freq = int(round(5_000 * freq_factor, 0))
        # freq = 5_000 -> 5/5_000 = 1ms pro count
        self.sm = StateMachine(1, wave_train_asym,
                                   freq=freq,
                                   sideset_base=Pin(output_pin))
        self.sm.restart()

    def transmit(self, pulse_cnt, pulse_high, pulse_low):
        # Angabe der Zeiten in Millisekunden (wenn freq_factor = 1)
        self.sm.put(pulse_cnt)
        self.sm.put(pulse_high - 1)
        self.sm.put(pulse_low - 1)
 
    def activate(self, state=1):
        self.sm.active(1 if state else 0)


if __name__ == "__main__":
    
    output_pin = 16
    
    wave_cnt = 2  
    high_cnt = 9
    low_cnt = 1
    
    wav = Wave_Train(output_pin=output_pin, freq_factor=1)
    wav.activate()
    wav.transmit(pulse_cnt = wave_cnt, pulse_high=high_cnt, pulse_low=low_cnt)
        
    time.sleep(.3)
    wav.activate(0)
    print("\ndone")